from .. import api
from ..api import request
from ..context import ctx
from ..objects import File
from ..objects.tg_methods import GetFile


def get_file(
        file_id: str = None,
) -> File:
    return request(
        GetFile,
        locals(),
        file_id=ctx.file_id,
    )


def get_file_url(
        file_id: str = None,
) -> str:
    file = get_file(file_id)
    return api.get_file_url(file.file_path)



from ..api import request
from ..objects import BotCommand, BotCommandScope
from ..objects.tg_methods import SetMyCommands


def set_my_commands(
        commands: list[BotCommand],
        scope: BotCommandScope = None,
        language_code: str = None,
) -> bool:
    return request(
        SetMyCommands,
        locals(),
    )



def clean_dict(value: dict) -> dict:
    from .cast import BaseModel, cast

    """ Recursively clean dict from `None` values """

    def clean_obj(obj):
        if isinstance(obj, BaseModel):
            return cast(obj, dict)
        if isinstance(obj, list):
            return [clean_obj(i) for i in obj]
        if isinstance(obj, dict):
            return {k: clean_obj(v) for k, v in obj.items() if v is not None}
        return obj

    return clean_obj(value)



from .base import ReplyMarkupT
from ..api import request
from ..context import ctx
from ..objects import MessageEntity, Message, InputFile
from ..objects.tg_methods import SendPhoto


def send_photo(
        photo: InputFile | str,
        reply_markup: ReplyMarkupT = None,
        caption: str = None,

        chat_id: int | str = None,
        parse_mode: str = None,
        disable_notification: bool = None,
        protect_content: bool = None,

        caption_entities: list[MessageEntity] = None,
        reply_to_message_id: int = None,
        allow_sending_without_reply: bool = None,
) -> Message:
    return request(
        SendPhoto,
        locals(),
        chat_id=ctx.chat_id,
        parse_mode=ctx.parse_mode,
        disable_notification=ctx.disable_notification,
        protect_content=ctx.protect_content,
    )



from __future__ import annotations

from dataclasses import dataclass, field
from hashlib import md5

import mongoengine as me

from ..new.translations import Translations
from ...base import BaseDocument, NewObject


class CallbackButtonDoc(BaseDocument):
    hash: str = me.StringField(primary_key=True)
    text: str = me.StringField()
    button_id: str = me.StringField()
    vars: dict = me.DictField()

    meta = {
        'collection': 'CallbackButtons'
    }


@dataclass
class CallbackButton(NewObject):
    """ CallbackButton """
    text: str | Translations
    id: str = None
    vars: dict = field(default_factory=dict)

    def __getitem__(self, item):
        return self.vars[item]

    def get_hash(self) -> str:
        string = f'{self.text}|{self.id}|{self.vars}'
        return md5(string.encode()).hexdigest()

    @classmethod
    def get(cls, doc_id: str) -> CallbackButton | None:
        doc = CallbackButtonDoc.get_doc(hash=doc_id)
        if not doc:
            return None
        return CallbackButton(text=doc.text, id=doc.button_id, vars=doc.vars)

    def __post_init__(self):
        self.id = self.id or str(self.text)
        btn_hash = self.get_hash()

        CallbackButtonDoc(
            hash=btn_hash,
            text=str(self.text),
            button_id=self.id,
            vars=self.vars,
        ).save()

        self.doc_id = btn_hash

    def __call__(self, **_vars) -> CallbackButton:
        """ Return button copy [text=text.format(**_vars), id=id.format(**_vars), vars=_vars] """
        return CallbackButton(self.text.format(**_vars), self.id, vars=_vars)



from ..api import request
from ..context import ctx
from ..objects import Chat
from ..objects.tg_methods import GetChat


def get_chat(
        chat_id: int | str = None,
) -> Chat:
    return request(
        GetChat,
        locals(),
        chat_id=ctx.chat_id,
    )



from ..api import request
from ..objects import BotCommandScope
from ..objects.tg_methods import DeleteMyCommands


def delete_my_commands(
        scope: BotCommandScope = None,
        language_code: str = None,
) -> bool:
    return request(
        DeleteMyCommands,
        locals(),
    )



from __future__ import annotations

import typing

import mongoengine as me

from .base_document import BaseDocument
from .base_model import BaseModel
from ..utils import cast

ObjectT = typing.TypeVar('ObjectT', bound='MyObject')


def get_model_name(obj: BaseModel | type[BaseModel]):
    if isinstance(obj, BaseModel):
        return obj.__class__.__name__
    return obj.__name__


class CurrentObject(BaseDocument):
    user_id: int = me.IntField()
    model: str = me.StringField()
    obj: dict = me.DictField()

    meta = {
        'collection': 'CurrentObjects',
    }


class CollectionObject(BaseDocument):
    user_id: int = me.IntField()
    model: str = me.StringField()
    obj: dict = me.DictField()

    meta = {
        'collection': 'CollectionObjects',
    }


class UserProxy(typing.Generic[ObjectT]):

    def __init__(self, obj: ObjectT):
        self.obj = obj

    def __enter__(self) -> ObjectT:
        return self.obj

    def __exit__(self, exc_type, exc_value, trace):
        self.obj.save()


class UserProxyModel(BaseModel):
    """ Base class for all project objects """

    @classmethod
    def get(cls: type[ObjectT]) -> ObjectT | None:
        """ Get current object of current user from storage """
        from ..context import ctx

        doc = CurrentObject.get_doc(user_id=ctx.user_id, model=get_model_name(cls))

        if doc is None:
            return None
        else:
            return cast(doc.obj, cls)

    @classmethod
    def delete(cls: type[ObjectT]) -> None:
        """ Delete current object of current user from storage """
        from ..context import ctx

        doc = CurrentObject.get_doc(user_id=ctx.user_id, model=get_model_name(cls))

        if doc:
            doc.delete()

    def save(self: ObjectT) -> ObjectT:
        """ Save object to storage as current for current user """
        from ..context import ctx

        doc = CurrentObject.get_doc(user_id=ctx.user_id, model=get_model_name(self))

        if doc is None:
            doc = CurrentObject(user_id=ctx.user_id, model=get_model_name(self))

        doc.obj = cast(self, dict)
        doc.save()

        return self

    @classmethod
    def proxy(cls: type[ObjectT]) -> UserProxy[ObjectT]:
        """
        (ContextManager)
        On enter: return newly created or current object of current user from storage;
        On exit: save object to storage as current for current user;
        """
        return UserProxy(cls.get() or cls())

    @classmethod
    def get_collection(cls: type[ObjectT]) -> list[ObjectT]:
        """ Get objects collection of current user from storage """
        from ..context import ctx

        docs = CollectionObject.get_docs(user_id=ctx.user_id, model=get_model_name(cls))
        return cast([doc.obj for doc in docs], list[cls])

    def save_to_collection(self: ObjectT) -> ObjectT:
        """ Save object to storage collection for current user """
        from ..context import ctx

        CollectionObject(
            user_id=ctx.user_id,
            model=get_model_name(self),
            obj=cast(self, dict)
        ).save()

        return self

    def replace_in_collection(self: ObjectT, key: str) -> ObjectT:
        """ Replace object in storage collection for current user by given key """
        from ..context import ctx

        docs = CollectionObject.get_docs(user_id=ctx.user_id, model=get_model_name(self))

        for doc in docs:
            if doc.obj[key] == getattr(self, key):
                doc.obj = cast(self, dict)
                doc.save()

        return self



from .base_document import BaseDocument
from .base_model import BaseModel
from .new_object import NewObject
from .tg_method import TgMethod
from .tg_object import TgObject
from .constants import Constants
from .user_proxy_model import UserProxyModel
from . import exceptions



""" All telegram objects from https://core.telegram.org/bots/api (v 6.0) """

from __future__ import annotations

from dataclasses import dataclass

from ..base import TgObject


# ==> Section: https://core.telegram.org/bots/api#getting-updates

@dataclass
class Update(TgObject):
    """ https://core.telegram.org/bots/api#update """

    update_id: int
    message: Message = None
    edited_message: Message = None
    channel_post: Message = None
    edited_channel_post: Message = None
    inline_query: InlineQuery = None
    chosen_inline_result: ChosenInlineResult = None
    callback_query: CallbackQuery = None
    shipping_query: ShippingQuery = None
    pre_checkout_query: PreCheckoutQuery = None
    poll: Poll = None
    poll_answer: PollAnswer = None
    my_chat_member: ChatMemberUpdated = None
    chat_member: ChatMemberUpdated = None
    chat_join_request: ChatJoinRequest = None


@dataclass
class WebhookInfo(TgObject):
    """ https://core.telegram.org/bots/api#webhookinfo """

    url: str
    has_custom_certificate: bool
    pending_update_count: int
    ip_address: str = None
    last_error_date: int = None
    last_error_message: str = None
    last_synchronization_error_date: int = None
    max_connections: int = None
    allowed_updates: list[str] = None


# ==> Section: https://core.telegram.org/bots/api#available-types

@dataclass
class User(TgObject):
    """ https://core.telegram.org/bots/api#user """

    id: int
    is_bot: bool
    first_name: str
    last_name: str = None
    username: str = None
    language_code: str = None
    can_join_groups: bool = None
    can_read_all_group_messages: bool = None
    supports_inline_queries: bool = None


@dataclass
class Chat(TgObject):
    """ https://core.telegram.org/bots/api#chat """

    id: int
    type: str
    title: str = None
    username: str = None
    first_name: str = None
    last_name: str = None
    photo: ChatPhoto = None
    bio: str = None
    has_private_forwards: bool = None
    description: str = None
    invite_link: str = None
    pinned_message: Message = None
    permissions: ChatPermissions = None
    slow_mode_delay: int = None
    message_auto_delete_time: int = None
    has_protected_content: bool = None
    sticker_set_name: str = None
    can_set_sticker_set: bool = None
    linked_chat_id: int = None
    location: ChatLocation = None


@dataclass
class Message(TgObject):
    """ https://core.telegram.org/bots/api#message """

    message_id: int
    date: int
    chat: Chat
    from_user: User = None
    sender_chat: Chat = None
    forward_from: User = None
    forward_from_chat: Chat = None
    forward_from_message_id: int = None
    forward_signature: str = None
    forward_sender_name: str = None
    forward_date: int = None
    is_automatic_forward: bool = None
    reply_to_message: Message = None
    via_bot: User = None
    edit_date: int = None
    has_protected_content: bool = None
    media_group_id: str = None
    author_signature: str = None
    text: str = None
    entities: list[MessageEntity] = None
    animation: Animation = None
    audio: Audio = None
    document: Document = None
    photo: list[PhotoSize] = None
    sticker: Sticker = None
    video: Video = None
    video_note: VideoNote = None
    voice: Voice = None
    caption: str = None
    caption_entities: list[MessageEntity] = None
    contact: Contact = None
    dice: Dice = None
    game: Game = None
    poll: Poll = None
    venue: Venue = None
    location: Location = None
    new_chat_members: list[User] = None
    left_chat_member: User = None
    new_chat_title: str = None
    new_chat_photo: list[PhotoSize] = None
    delete_chat_photo: bool = None
    group_chat_created: bool = None
    supergroup_chat_created: bool = None
    channel_chat_created: bool = None
    message_auto_delete_timer_changed: MessageAutoDeleteTimerChanged = None
    migrate_to_chat_id: int = None
    migrate_from_chat_id: int = None
    pinned_message: Message = None
    invoice: Invoice = None
    successful_payment: SuccessfulPayment = None
    connected_website: str = None
    passport_data: PassportData = None
    proximity_alert_triggered: ProximityAlertTriggered = None
    video_chat_scheduled: VideoChatScheduled = None
    video_chat_started: VideoChatStarted = None
    video_chat_ended: VideoChatEnded = None
    video_chat_participants_invited: VideoChatParticipantsInvited = None
    web_app_data: WebAppData = None
    reply_markup: InlineKeyboardMarkup = None


@dataclass
class MessageId(TgObject):
    """ https://core.telegram.org/bots/api#messageid """

    message_id: int


@dataclass
class MessageEntity(TgObject):
    """ https://core.telegram.org/bots/api#messageentity """

    type: str
    offset: int
    length: int
    url: str = None
    user: User = None
    language: str = None


@dataclass
class PhotoSize(TgObject):
    """ https://core.telegram.org/bots/api#photosize """

    file_id: str
    file_unique_id: str
    width: int
    height: int
    file_size: int = None


@dataclass
class Animation(TgObject):
    """ https://core.telegram.org/bots/api#animation """

    file_id: str
    file_unique_id: str
    width: int
    height: int
    duration: int
    thumb: PhotoSize = None
    file_name: str = None
    mime_type: str = None
    file_size: int = None


@dataclass
class Audio(TgObject):
    """ https://core.telegram.org/bots/api#audio """

    file_id: str
    file_unique_id: str
    duration: int
    performer: str = None
    title: str = None
    file_name: str = None
    mime_type: str = None
    file_size: int = None
    thumb: PhotoSize = None


@dataclass
class Document(TgObject):
    """ https://core.telegram.org/bots/api#document """

    file_id: str
    file_unique_id: str
    thumb: PhotoSize = None
    file_name: str = None
    mime_type: str = None
    file_size: int = None


@dataclass
class Video(TgObject):
    """ https://core.telegram.org/bots/api#video """

    file_id: str
    file_unique_id: str
    width: int
    height: int
    duration: int
    thumb: PhotoSize = None
    file_name: str = None
    mime_type: str = None
    file_size: int = None


@dataclass
class VideoNote(TgObject):
    """ https://core.telegram.org/bots/api#videonote """

    file_id: str
    file_unique_id: str
    length: int
    duration: int
    thumb: PhotoSize = None
    file_size: int = None


@dataclass
class Voice(TgObject):
    """ https://core.telegram.org/bots/api#voice """

    file_id: str
    file_unique_id: str
    duration: int
    mime_type: str = None
    file_size: int = None


@dataclass
class Contact(TgObject):
    """ https://core.telegram.org/bots/api#contact """

    phone_number: str
    first_name: str
    last_name: str = None
    user_id: int = None
    vcard: str = None


@dataclass
class Dice(TgObject):
    """ https://core.telegram.org/bots/api#dice """

    emoji: str
    value: int


@dataclass
class PollOption(TgObject):
    """ https://core.telegram.org/bots/api#polloption """

    text: str
    voter_count: int


@dataclass
class PollAnswer(TgObject):
    """ https://core.telegram.org/bots/api#pollanswer """

    poll_id: str
    user: User
    option_ids: list[int]


@dataclass
class Poll(TgObject):
    """ https://core.telegram.org/bots/api#poll """

    id: str
    question: str
    options: list[PollOption]
    total_voter_count: int
    is_closed: bool
    is_anonymous: bool
    type: str
    allows_multiple_answers: bool
    correct_option_id: int = None
    explanation: str = None
    explanation_entities: list[MessageEntity] = None
    open_period: int = None
    close_date: int = None


@dataclass
class Location(TgObject):
    """ https://core.telegram.org/bots/api#location """

    longitude: float
    latitude: float
    horizontal_accuracy: float = None
    live_period: int = None
    heading: int = None
    proximity_alert_radius: int = None


@dataclass
class Venue(TgObject):
    """ https://core.telegram.org/bots/api#venue """

    location: Location
    title: str
    address: str
    foursquare_id: str = None
    foursquare_type: str = None
    google_place_id: str = None
    google_place_type: str = None


@dataclass
class WebAppData(TgObject):
    """ https://core.telegram.org/bots/api#webappdata """

    data: str
    button_text: str


@dataclass
class ProximityAlertTriggered(TgObject):
    """ https://core.telegram.org/bots/api#proximityalerttriggered """

    traveler: User
    watcher: User
    distance: int


@dataclass
class MessageAutoDeleteTimerChanged(TgObject):
    """ https://core.telegram.org/bots/api#messageautodeletetimerchanged """

    message_auto_delete_time: int


@dataclass
class VideoChatScheduled(TgObject):
    """ https://core.telegram.org/bots/api#videochatscheduled """

    start_date: int


@dataclass
class VideoChatStarted(TgObject):
    """ https://core.telegram.org/bots/api#videochatstarted """


@dataclass
class VideoChatEnded(TgObject):
    """ https://core.telegram.org/bots/api#videochatended """

    duration: int


@dataclass
class VideoChatParticipantsInvited(TgObject):
    """ https://core.telegram.org/bots/api#videochatparticipantsinvited """

    users: list[User]


@dataclass
class UserProfilePhotos(TgObject):
    """ https://core.telegram.org/bots/api#userprofilephotos """

    total_count: int
    photos: list[list[PhotoSize]]


@dataclass
class File(TgObject):
    """ https://core.telegram.org/bots/api#file """

    file_id: str
    file_unique_id: str
    file_size: int = None
    file_path: str = None


@dataclass
class WebAppInfo(TgObject):
    """ https://core.telegram.org/bots/api#webappinfo """

    url: str


@dataclass
class ReplyKeyboardMarkup(TgObject):
    """ https://core.telegram.org/bots/api#replykeyboardmarkup """

    keyboard: list[list[KeyboardButton]]
    resize_keyboard: bool = None
    one_time_keyboard: bool = None
    input_field_placeholder: str = None
    selective: bool = None


@dataclass
class KeyboardButton(TgObject):
    """ https://core.telegram.org/bots/api#keyboardbutton """

    text: str
    request_contact: bool = None
    request_location: bool = None
    request_poll: KeyboardButtonPollType = None
    web_app: WebAppInfo = None


@dataclass
class KeyboardButtonPollType(TgObject):
    """ https://core.telegram.org/bots/api#keyboardbuttonpolltype """

    type: str = None


@dataclass
class ReplyKeyboardRemove(TgObject):
    """ https://core.telegram.org/bots/api#replykeyboardremove """

    remove_keyboard: bool = True
    selective: bool = None


@dataclass
class InlineKeyboardMarkup(TgObject):
    """ https://core.telegram.org/bots/api#inlinekeyboardmarkup """

    inline_keyboard: list[list[InlineKeyboardButton]]


@dataclass
class InlineKeyboardButton(TgObject):
    """ https://core.telegram.org/bots/api#inlinekeyboardbutton """

    text: str
    url: str = None
    callback_data: str = None
    web_app: WebAppInfo = None
    login_url: LoginUrl = None
    switch_inline_query: str = None
    switch_inline_query_current_chat: str = None
    callback_game: CallbackGame = None
    pay: bool = None


@dataclass
class LoginUrl(TgObject):
    """ https://core.telegram.org/bots/api#loginurl """

    url: str
    forward_text: str = None
    bot_username: str = None
    request_write_access: bool = None


@dataclass
class CallbackQuery(TgObject):
    """ https://core.telegram.org/bots/api#callbackquery """

    id: str
    from_user: User
    chat_instance: str
    message: Message = None
    inline_message_id: str = None
    data: str = None
    game_short_name: str = None


@dataclass
class ForceReply(TgObject):
    """ https://core.telegram.org/bots/api#forcereply """

    force_reply: bool
    input_field_placeholder: str = None
    selective: bool = None


@dataclass
class ChatPhoto(TgObject):
    """ https://core.telegram.org/bots/api#chatphoto """

    small_file_id: str
    small_file_unique_id: str
    big_file_id: str
    big_file_unique_id: str


@dataclass
class ChatInviteLink(TgObject):
    """ https://core.telegram.org/bots/api#chatinvitelink """

    invite_link: str
    creator: User
    creates_join_request: bool
    is_primary: bool
    is_revoked: bool
    name: str = None
    expire_date: int = None
    member_limit: int = None
    pending_join_request_count: int = None


@dataclass
class ChatAdministratorRights(TgObject):
    """ https://core.telegram.org/bots/api#chatadministratorrights """

    is_anonymous: bool
    can_manage_chat: bool
    can_delete_messages: bool
    can_manage_video_chats: bool
    can_restrict_members: bool
    can_promote_members: bool
    can_change_info: bool
    can_invite_users: bool
    can_post_messages: bool = None
    can_edit_messages: bool = None
    can_pin_messages: bool = None


# abstract
class ChatMember(TgObject):
    """ https://core.telegram.org/bots/api#chatmember """

    status: str
    user: User


@dataclass
class ChatMemberOwner(ChatMember):
    """ https://core.telegram.org/bots/api#chatmemberowner """

    user: User
    is_anonymous: bool
    custom_title: str = None
    status: str = 'creator'


@dataclass
class ChatMemberAdministrator(ChatMember):
    """ https://core.telegram.org/bots/api#chatmemberadministrator """

    user: User
    can_be_edited: bool
    is_anonymous: bool
    can_manage_chat: bool
    can_delete_messages: bool
    can_manage_video_chats: bool
    can_restrict_members: bool
    can_promote_members: bool
    can_change_info: bool
    can_invite_users: bool
    can_post_messages: bool = None
    can_edit_messages: bool = None
    can_pin_messages: bool = None
    custom_title: str = None
    status: str = 'administrator'


@dataclass
class ChatMemberMember(ChatMember):
    """ https://core.telegram.org/bots/api#chatmembermember """

    user: User
    status: str = 'member'


@dataclass
class ChatMemberRestricted(ChatMember):
    """ https://core.telegram.org/bots/api#chatmemberrestricted """

    user: User
    is_member: bool
    can_change_info: bool
    can_invite_users: bool
    can_pin_messages: bool
    can_send_messages: bool
    can_send_media_messages: bool
    can_send_polls: bool
    can_send_other_messages: bool
    can_add_web_page_previews: bool
    until_date: int
    status: str = 'restricted'


@dataclass
class ChatMemberLeft(ChatMember):
    """ https://core.telegram.org/bots/api#chatmemberleft """

    user: User
    status: str = 'left'


@dataclass
class ChatMemberBanned(ChatMember):
    """ https://core.telegram.org/bots/api#chatmemberbanned """

    user: User
    until_date: int
    status: str = 'kicked'


ChatMemberT = ChatMemberAdministrator | ChatMemberOwner | ChatMemberRestricted \
              | ChatMemberBanned | ChatMemberMember | ChatMemberLeft


@dataclass
class ChatMemberUpdated(TgObject):
    """ https://core.telegram.org/bots/api#chatmemberupdated """

    chat: Chat
    from_user: User
    date: int
    old_chat_member: ChatMemberT
    new_chat_member: ChatMemberT
    invite_link: ChatInviteLink = None


@dataclass
class ChatJoinRequest(TgObject):
    """ https://core.telegram.org/bots/api#chatjoinrequest """

    chat: Chat
    from_user: User
    date: int
    bio: str = None
    invite_link: ChatInviteLink = None


@dataclass
class ChatPermissions(TgObject):
    """ https://core.telegram.org/bots/api#chatpermissions """

    can_send_messages: bool = None
    can_send_media_messages: bool = None
    can_send_polls: bool = None
    can_send_other_messages: bool = None
    can_add_web_page_previews: bool = None
    can_change_info: bool = None
    can_invite_users: bool = None
    can_pin_messages: bool = None


@dataclass
class ChatLocation(TgObject):
    """ https://core.telegram.org/bots/api#chatlocation """

    location: Location
    address: str


@dataclass
class BotCommand(TgObject):
    """ https://core.telegram.org/bots/api#botcommand """

    command: str
    description: str


# abstract
class BotCommandScope(TgObject):
    """ https://core.telegram.org/bots/api#botcommandscope """

    type: str


@dataclass
class BotCommandScopeDefault(BotCommandScope):
    """ https://core.telegram.org/bots/api#botcommandscopedefault """

    type: str = 'default'


@dataclass
class BotCommandScopeAllPrivateChats(BotCommandScope):
    """ https://core.telegram.org/bots/api#botcommandscopeallprivatechats """

    type: str = 'all_private_chats'


@dataclass
class BotCommandScopeAllGroupChats(BotCommandScope):
    """ https://core.telegram.org/bots/api#botcommandscopeallgroupchats """

    type: str = 'all_group_chats'


@dataclass
class BotCommandScopeAllChatAdministrators(BotCommandScope):
    """ https://core.telegram.org/bots/api#botcommandscopeallchatadministrators """

    type: str = 'all_chat_administrators'


@dataclass
class BotCommandScopeChat(BotCommandScope):
    """ https://core.telegram.org/bots/api#botcommandscopechat """

    chat_id: int | str
    type: str = 'chat'


@dataclass
class BotCommandScopeChatAdministrators(BotCommandScope):
    """ https://core.telegram.org/bots/api#botcommandscopechatadministrators """

    chat_id: int | str
    type: str = 'chat_administrators'


@dataclass
class BotCommandScopeChatMember(BotCommandScope):
    """ https://core.telegram.org/bots/api#botcommandscopechatmember """

    chat_id: int | str
    user_id: int
    type: str = 'chat_member'


@dataclass
class MenuButton(TgObject):
    """ https://core.telegram.org/bots/api#menubutton """


@dataclass
class MenuButtonCommands(MenuButton):
    """ https://core.telegram.org/bots/api#menubuttoncommands """

    type: str


@dataclass
class MenuButtonWebApp(MenuButton):
    """ https://core.telegram.org/bots/api#menubuttonwebapp """

    type: str
    text: str
    web_app: WebAppInfo


@dataclass
class MenuButtonDefault(MenuButton):
    """ https://core.telegram.org/bots/api#menubuttondefault """

    type: str


@dataclass
class ResponseParameters(TgObject):
    """ https://core.telegram.org/bots/api#responseparameters """

    migrate_to_chat_id: int = None
    retry_after: int = None


@dataclass
class InputMedia(TgObject):
    """ https://core.telegram.org/bots/api#inputmedia """


@dataclass
class InputMediaPhoto(InputMedia):
    """ https://core.telegram.org/bots/api#inputmediaphoto """

    media: str
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    type: str = 'photo'


@dataclass
class InputMediaVideo(InputMedia):
    """ https://core.telegram.org/bots/api#inputmediavideo """

    media: str
    thumb: InputFile | str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    width: int = None
    height: int = None
    duration: int = None
    supports_streaming: bool = None
    type: str = 'video'


@dataclass
class InputMediaAnimation(InputMedia):
    """ https://core.telegram.org/bots/api#inputmediaanimation """

    media: str
    thumb: InputFile | str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    width: int = None
    height: int = None
    duration: int = None
    type: str = 'animation'


@dataclass
class InputMediaAudio(InputMedia):
    """ https://core.telegram.org/bots/api#inputmediaaudio """

    media: str
    thumb: InputFile | str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    duration: int = None
    performer: str = None
    title: str = None
    type: str = 'audio'


@dataclass
class InputMediaDocument(InputMedia):
    """ https://core.telegram.org/bots/api#inputmediadocument """

    media: str
    thumb: InputFile | str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    disable_content_type_detection: bool = None
    type: str = 'document'


@dataclass
class InputFile(TgObject):
    """ https://core.telegram.org/bots/api#inputfile """

    path: str


# ==> Section: https://core.telegram.org/bots/api#stickers

@dataclass
class Sticker(TgObject):
    """ https://core.telegram.org/bots/api#sticker """

    file_id: str
    file_unique_id: str
    width: int
    height: int
    is_animated: bool
    is_video: bool
    thumb: PhotoSize = None
    emoji: str = None
    set_name: str = None
    mask_position: MaskPosition = None
    file_size: int = None


@dataclass
class StickerSet(TgObject):
    """ https://core.telegram.org/bots/api#stickerset """

    name: str
    title: str
    is_animated: bool
    is_video: bool
    contains_masks: bool
    stickers: list[Sticker]
    thumb: PhotoSize = None


@dataclass
class MaskPosition(TgObject):
    """ https://core.telegram.org/bots/api#maskposition """

    point: str
    x_shift: float
    y_shift: float
    scale: float


# ==> Section: https://core.telegram.org/bots/api#inline-mode

@dataclass
class InlineQuery(TgObject):
    """ https://core.telegram.org/bots/api#inlinequery """

    id: str
    from_user: User
    query: str
    offset: str
    chat_type: str = None
    location: Location = None


@dataclass
class InlineQueryResult(TgObject):
    """ https://core.telegram.org/bots/api#inlinequeryresult """


@dataclass
class InlineQueryResultArticle(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultarticle """

    type: str
    id: str
    title: str
    input_message_content: InputMessageContent
    reply_markup: InlineKeyboardMarkup = None
    url: str = None
    hide_url: bool = None
    description: str = None
    thumb_url: str = None
    thumb_width: int = None
    thumb_height: int = None


@dataclass
class InlineQueryResultPhoto(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultphoto """

    type: str
    id: str
    photo_url: str
    thumb_url: str
    photo_width: int = None
    photo_height: int = None
    title: str = None
    description: str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InlineQueryResultGif(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultgif """

    type: str
    id: str
    gif_url: str
    thumb_url: str
    gif_width: int = None
    gif_height: int = None
    gif_duration: int = None
    thumb_mime_type: str = None
    title: str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InlineQueryResultMpeg4Gif(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultmpeg4gif """

    type: str
    id: str
    mpeg4_url: str
    thumb_url: str
    mpeg4_width: int = None
    mpeg4_height: int = None
    mpeg4_duration: int = None
    thumb_mime_type: str = None
    title: str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InlineQueryResultVideo(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultvideo """

    type: str
    id: str
    video_url: str
    mime_type: str
    thumb_url: str
    title: str
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    video_width: int = None
    video_height: int = None
    video_duration: int = None
    description: str = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InlineQueryResultAudio(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultaudio """

    type: str
    id: str
    audio_url: str
    title: str
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    performer: str = None
    audio_duration: int = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InlineQueryResultVoice(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultvoice """

    type: str
    id: str
    voice_url: str
    title: str
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    voice_duration: int = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InlineQueryResultDocument(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultdocument """

    type: str
    id: str
    title: str
    document_url: str
    mime_type: str
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    description: str = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None
    thumb_url: str = None
    thumb_width: int = None
    thumb_height: int = None


@dataclass
class InlineQueryResultLocation(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultlocation """

    type: str
    id: str
    latitude: float
    longitude: float
    title: str
    horizontal_accuracy: float = None
    live_period: int = None
    heading: int = None
    proximity_alert_radius: int = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None
    thumb_url: str = None
    thumb_width: int = None
    thumb_height: int = None


@dataclass
class InlineQueryResultVenue(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultvenue """

    type: str
    id: str
    latitude: float
    longitude: float
    title: str
    address: str
    foursquare_id: str = None
    foursquare_type: str = None
    google_place_id: str = None
    google_place_type: str = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None
    thumb_url: str = None
    thumb_width: int = None
    thumb_height: int = None


@dataclass
class InlineQueryResultContact(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultcontact """

    type: str
    id: str
    phone_number: str
    first_name: str
    last_name: str = None
    vcard: str = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None
    thumb_url: str = None
    thumb_width: int = None
    thumb_height: int = None


@dataclass
class InlineQueryResultGame(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultgame """

    type: str
    id: str
    game_short_name: str
    reply_markup: InlineKeyboardMarkup = None


@dataclass
class InlineQueryResultCachedPhoto(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultcachedphoto """

    type: str
    id: str
    photo_file_id: str
    title: str = None
    description: str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InlineQueryResultCachedGif(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultcachedgif """

    type: str
    id: str
    gif_file_id: str
    title: str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InlineQueryResultCachedMpeg4Gif(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultcachedmpeg4gif """

    type: str
    id: str
    mpeg4_file_id: str
    title: str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InlineQueryResultCachedSticker(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultcachedsticker """

    type: str
    id: str
    sticker_file_id: str
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InlineQueryResultCachedDocument(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultcacheddocument """

    type: str
    id: str
    title: str
    document_file_id: str
    description: str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InlineQueryResultCachedVideo(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultcachedvideo """

    type: str
    id: str
    video_file_id: str
    title: str
    description: str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InlineQueryResultCachedVoice(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultcachedvoice """

    type: str
    id: str
    voice_file_id: str
    title: str
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InlineQueryResultCachedAudio(InlineQueryResult):
    """ https://core.telegram.org/bots/api#inlinequeryresultcachedaudio """

    type: str
    id: str
    audio_file_id: str
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    reply_markup: InlineKeyboardMarkup = None
    input_message_content: InputMessageContent = None


@dataclass
class InputMessageContent(TgObject):
    """ https://core.telegram.org/bots/api#inputmessagecontent """


@dataclass
class InputTextMessageContent(InputMessageContent):
    """ https://core.telegram.org/bots/api#inputtextmessagecontent """

    message_text: str
    parse_mode: str = None
    entities: list[MessageEntity] = None
    disable_web_page_preview: bool = None


@dataclass
class InputLocationMessageContent(InputMessageContent):
    """ https://core.telegram.org/bots/api#inputlocationmessagecontent """

    latitude: float
    longitude: float
    horizontal_accuracy: float = None
    live_period: int = None
    heading: int = None
    proximity_alert_radius: int = None


@dataclass
class InputVenueMessageContent(InputMessageContent):
    """ https://core.telegram.org/bots/api#inputvenuemessagecontent """

    latitude: float
    longitude: float
    title: str
    address: str
    foursquare_id: str = None
    foursquare_type: str = None
    google_place_id: str = None
    google_place_type: str = None


@dataclass
class InputContactMessageContent(InputMessageContent):
    """ https://core.telegram.org/bots/api#inputcontactmessagecontent """

    phone_number: str
    first_name: str
    last_name: str = None
    vcard: str = None


@dataclass
class InputInvoiceMessageContent(InputMessageContent):
    """ https://core.telegram.org/bots/api#inputinvoicemessagecontent """

    title: str
    description: str
    payload: str
    provider_token: str
    currency: str
    prices: list[LabeledPrice]
    max_tip_amount: int = None
    suggested_tip_amounts: list[int] = None
    provider_data: str = None
    photo_url: str = None
    photo_size: int = None
    photo_width: int = None
    photo_height: int = None
    need_name: bool = None
    need_phone_number: bool = None
    need_email: bool = None
    need_shipping_address: bool = None
    send_phone_number_to_provider: bool = None
    send_email_to_provider: bool = None
    is_flexible: bool = None


@dataclass
class ChosenInlineResult(TgObject):
    """ https://core.telegram.org/bots/api#choseninlineresult """

    result_id: str
    from_user: User
    query: str
    location: Location = None
    inline_message_id: str = None


@dataclass
class SentWebAppMessage(TgObject):
    """ https://core.telegram.org/bots/api#sentwebappmessage """

    inline_message_id: str = None


# ==> Section: https://core.telegram.org/bots/api#payments

@dataclass
class LabeledPrice(TgObject):
    """ https://core.telegram.org/bots/api#labeledprice """

    label: str
    amount: int


@dataclass
class Invoice(TgObject):
    """ https://core.telegram.org/bots/api#invoice """

    title: str
    description: str
    start_parameter: str
    currency: str
    total_amount: int


@dataclass
class ShippingAddress(TgObject):
    """ https://core.telegram.org/bots/api#shippingaddress """

    country_code: str
    state: str
    city: str
    street_line1: str
    street_line2: str
    post_code: str


@dataclass
class OrderInfo(TgObject):
    """ https://core.telegram.org/bots/api#orderinfo """

    name: str = None
    phone_number: str = None
    email: str = None
    shipping_address: ShippingAddress = None


@dataclass
class ShippingOption(TgObject):
    """ https://core.telegram.org/bots/api#shippingoption """

    id: str
    title: str
    prices: list[LabeledPrice]


@dataclass
class SuccessfulPayment(TgObject):
    """ https://core.telegram.org/bots/api#successfulpayment """

    currency: str
    total_amount: int
    invoice_payload: str
    telegram_payment_charge_id: str
    provider_payment_charge_id: str
    shipping_option_id: str = None
    order_info: OrderInfo = None


@dataclass
class ShippingQuery(TgObject):
    """ https://core.telegram.org/bots/api#shippingquery """

    id: str
    from_user: User
    invoice_payload: str
    shipping_address: ShippingAddress


@dataclass
class PreCheckoutQuery(TgObject):
    """ https://core.telegram.org/bots/api#precheckoutquery """

    id: str
    from_user: User
    currency: str
    total_amount: int
    invoice_payload: str
    shipping_option_id: str = None
    order_info: OrderInfo = None


# ==> Section: https://core.telegram.org/bots/api#telegram-passport

@dataclass
class PassportData(TgObject):
    """ https://core.telegram.org/bots/api#passportdata """

    data: list[EncryptedPassportElement]
    credentials: EncryptedCredentials


@dataclass
class PassportFile(TgObject):
    """ https://core.telegram.org/bots/api#passportfile """

    file_id: str
    file_unique_id: str
    file_size: int
    file_date: int


@dataclass
class EncryptedPassportElement(TgObject):
    """ https://core.telegram.org/bots/api#encryptedpassportelement """

    type: str
    hash: str
    data: str = None
    phone_number: str = None
    email: str = None
    files: list[PassportFile] = None
    front_side: PassportFile = None
    reverse_side: PassportFile = None
    selfie: PassportFile = None
    translation: list[PassportFile] = None


@dataclass
class EncryptedCredentials(TgObject):
    """ https://core.telegram.org/bots/api#encryptedcredentials """

    data: str
    hash: str
    secret: str


@dataclass
class PassportElementError(TgObject):
    """ https://core.telegram.org/bots/api#passportelementerror """


@dataclass
class PassportElementErrorDataField(PassportElementError):
    """ https://core.telegram.org/bots/api#passportelementerrordatafield """

    source: str
    type: str
    field_name: str
    data_hash: str
    message: str


@dataclass
class PassportElementErrorFrontSide(PassportElementError):
    """ https://core.telegram.org/bots/api#passportelementerrorfrontside """

    source: str
    type: str
    file_hash: str
    message: str


@dataclass
class PassportElementErrorReverseSide(PassportElementError):
    """ https://core.telegram.org/bots/api#passportelementerrorreverseside """

    source: str
    type: str
    file_hash: str
    message: str


@dataclass
class PassportElementErrorSelfie(PassportElementError):
    """ https://core.telegram.org/bots/api#passportelementerrorselfie """

    source: str
    type: str
    file_hash: str
    message: str


@dataclass
class PassportElementErrorFile(PassportElementError):
    """ https://core.telegram.org/bots/api#passportelementerrorfile """

    source: str
    type: str
    file_hash: str
    message: str


@dataclass
class PassportElementErrorFiles(PassportElementError):
    """ https://core.telegram.org/bots/api#passportelementerrorfiles """

    source: str
    type: str
    file_hashes: list[str]
    message: str


@dataclass
class PassportElementErrorTranslationFile(PassportElementError):
    """ https://core.telegram.org/bots/api#passportelementerrortranslationfile """

    source: str
    type: str
    file_hash: str
    message: str


@dataclass
class PassportElementErrorTranslationFiles(PassportElementError):
    """ https://core.telegram.org/bots/api#passportelementerrortranslationfiles """

    source: str
    type: str
    file_hashes: list[str]
    message: str


@dataclass
class PassportElementErrorUnspecified(PassportElementError):
    """ https://core.telegram.org/bots/api#passportelementerrorunspecified """

    source: str
    type: str
    element_hash: str
    message: str


# ==> Section: https://core.telegram.org/bots/api#games

@dataclass
class Game(TgObject):
    """ https://core.telegram.org/bots/api#game """

    title: str
    description: str
    photo: list[PhotoSize]
    text: str = None
    text_entities: list[MessageEntity] = None
    animation: Animation = None


@dataclass
class CallbackGame(TgObject):
    """ https://core.telegram.org/bots/api#callbackgame """

    user_id: int
    score: int
    force: bool = None
    disable_edit_message: bool = None
    chat_id: int = None
    message_id: int = None
    inline_message_id: str = None


@dataclass
class GameHighScore(TgObject):
    """ https://core.telegram.org/bots/api#gamehighscore """

    position: int
    user: User
    score: int



import re
from dataclasses import dataclass

from .base import Filter
from ..context import ctx
from ..objects import CallbackButton, Translations
from ..utils import listify


@dataclass
class Button(Filter):
    value: CallbackButton | list[CallbackButton] | str | list[str] | Translations | list[Translations]

    def check_one(self, value: CallbackButton | str | Translations):
        if isinstance(value, CallbackButton):
            try:
                if not (doc_id := ctx.update.callback_query.data):
                    return False
            except AttributeError:
                return False

            btn = CallbackButton.get(doc_id)
            if not btn:
                return False
            return btn.id == self.value.id

        try:
            if not (text := ctx.update.message.text):
                return False
        except AttributeError:
            return False

        if isinstance(value, (str, Translations)):
            if isinstance(value, Translations):
                value = value.get()

            pattern = re.sub(r'\\\{.+}', '.+', re.escape(value))
            if re.fullmatch(pattern, text):
                return True

            return False

        raise ValueError('Invalid button')

    def __call__(self):
        for value in listify(self.value):
            if self.check_one(value):
                return True
        return False



def a(text, url) -> str:
    return f'<a href="{url}">{text}</a>'


def b(text) -> str:
    return f'<b>{text}</b>'


def spoiler(text) -> str:
    return f'<tg-spoiler>{text}</tg-spoiler>'


def code(text) -> str:
    return f'<code>{text}</code>'



from .layer1 import ContextLayer1
from ..objects import PhotoSize, Document


class ContextLayer2(ContextLayer1):

    @property
    def user_id(self) -> int | None:
        """ Message.user.id | CallbackQuery.user.id """
        value = None

        if user := self.user:
            value = user.id

        return value

    @property
    def query_data(self) -> str | None:
        """ CallbackQuery.data """

        value = None

        if callback_query := self.callback_query:
            value = callback_query.data

        return value

    @property
    def text(self) -> str | None:
        """ Message.text | ChannelPost.text | CallbackQuery.message.text """
        value = None

        if message := self.message:
            value = message.text

        return value

    @property
    def photo(self) -> PhotoSize | None:
        """ Message.photo[-1] | ChannelPost.photo[-1] | CallbackQuery.message.photo[-1] """
        if message := self.message:
            if photo := message.photo:
                return photo[-1]

    @property
    def document(self) -> Document | None:
        """ Message.document | ChannelPost.document | CallbackQuery.message.document """
        if message := self.message:
            return message.document

    @property
    def file_id(self) -> str | None:
        """
        Message.photo[-1].file_id | ChannelPost.photo[-1].file_id | CallbackQuery.message.photo[-1].file_id |
        Message.document.file_id | ChannelPost.document.file_id | CallbackQuery.message.document.file_id
        """
        if photo := self.photo:
            return photo.file_id
        if document := self.document:
            return document.file_id

    @property
    def chat_id(self) -> int | None:
        """ Message.chat.id | ChannelPost.chat.id | CallbackQuery.chat.id """
        value = None

        if chat := self.chat:
            value = chat.id

        return value

    @property
    def chat_type(self) -> int | None:
        """ Message.chat.type | ChannelPost.chat.type | CallbackQuery.chat.type """
        value = None

        if chat := self.chat:
            value = chat.type

        return value

    @property
    def callback_query_id(self) -> str | None:
        """ CallbackQuery.id """
        value = None

        if callback_query := self.callback_query:
            value = callback_query.id

        return value

    @property
    def message_id(self) -> int | None:
        """ Message.id | ChannelPost.id | CallbackQuery.message.id """
        value = None

        if message := self.message:
            value = message.message_id

        return value

    @property
    def media_group_id(self) -> str | None:
        """ Message.media_group_id | ChannelPost.media_group_id | CallbackQuery.message.media_group_id """
        value = None

        if message := self.message:
            value = message.media_group_id

        return value



from dataclasses import dataclass

from .base import Filter
from ..context import ctx
from ..utils import listify


@dataclass
class UserId(Filter):
    value: int | list[int]

    def __call__(self):
        return ctx.user_id in listify(self.value)



from .layer0 import ContextLayer0
from ..objects import Message, User, Chat, CallbackQuery, ChatMemberUpdated, ChatMemberT


class ContextLayer1(ContextLayer0):

    @property
    def message(self) -> Message | None:
        """ Message | ChannelPost | CallbackQuery.message """
        value = None

        if update := self.update:
            value = update.message or update.channel_post

            if value is None:
                if callback_query := self.callback_query:
                    value = callback_query.message

        return value

    @property
    def callback_query(self) -> CallbackQuery | None:
        """ CallbackQuery """
        value = None

        if update := self.update:
            value = update.callback_query

        return value

    @property
    def chat_member_updated(self) -> ChatMemberUpdated | None:
        """ MyChatMember | ChatMember """
        value = None

        if update := self.update:
            value = update.my_chat_member or update.chat_member

        return value

    @property
    def chat(self) -> Chat | None:
        """
        Message.chat | ChannelPost.chat | CallbackQuery.chat |
        MyChatMember.chat | ChatMember.chat
        """

        value = None

        if message := self.message:
            value = message.chat
        if chat_member_updated := self.chat_member_updated:
            value = chat_member_updated.chat

        return value

    @property
    def user(self) -> User | None:
        """ Message.user | CallbackQuery.user| MyChatMember.user | ChatMember.user """
        value = None

        if update := self.update:
            if message := update.message:
                value = message.from_user
            elif callback_query := update.callback_query:
                value = callback_query.from_user
            elif chat_member_updated := self.chat_member_updated:
                value = chat_member_updated.from_user

        return value

    @property
    def new_chat_member(self) -> ChatMemberT | None:
        """ MyChatMember.new_chat_member | ChatMember.new_chat_member """
        value = None

        if chat_member_updated := self.chat_member_updated:
            value = chat_member_updated.new_chat_member

        return value

    @property
    def old_chat_member(self) -> ChatMemberT | None:
        """ MyChatMember.old_chat_member | ChatMember.old_chat_member """
        value = None

        if chat_member_updated := self.chat_member_updated:
            value = chat_member_updated.old_chat_member

        return value



import typing
from dataclasses import dataclass, field

from ..filters.base import Filter


@dataclass
class Handler:
    func: typing.Callable[..., None]
    filters: list[Filter] = field(default_factory=list)
    exclusive: bool = True
    check_first: bool = False
    check_last: bool = False
    check_after_any: bool = False



from __future__ import annotations

import importlib
import types
import typing
from dataclasses import Field, asdict

from ..base import BaseModel

T = typing.TypeVar('T')


def is_generic_type(_type: type) -> bool:
    return hasattr(_type, '__origin__')


def get_type_origin(_type: type) -> type:
    return getattr(_type, '__origin__')


def is_union_type(_type) -> bool:
    return isinstance(_type, types.UnionType)


def get_type_args(_type: types.UnionType | type) -> list[type]:
    return list(getattr(_type, '__args__'))


def unify_type(_type: None | type | list[type] | types.UnionType) -> type | list[type]:
    if _type is None:
        return type(None)

    if is_union_type(_type):
        return [unify_type(subtype) for subtype in get_type_args(_type)]

    if isinstance(_type, list):
        return [unify_type(subtype) for subtype in _type]

    if is_generic_type(_type):
        if get_type_origin(_type) == list:
            return _type
        raise ValueError('Unsupported generic type')

    if isinstance(_type, type):
        return _type

    raise ValueError('Unsupported type')


assert unify_type(None) == type(None)
assert unify_type(str) == str
assert unify_type([str, None]) == [str, type(None)]
assert unify_type(str | None) == [str, type(None)]


def get_model_fields(_type: type[BaseModel]) -> list[Field]:
    return list(getattr(_type, '__dataclass_fields__').values())


def eval_field_type(_field: Field, cls: type[BaseModel]) -> type | types.UnionType:
    module = importlib.import_module(cls.__module__)
    return eval(_field.type, vars(module))


def cast_to_model(value: dict, to_type: type[BaseModel]):
    new_value = {}

    for _field in get_model_fields(to_type):
        if isinstance(_field.type, str):
            _field.type = eval_field_type(_field, cls=to_type)

        key = to_type.__aliases__.get(_field.name, _field.name)

        if key in value:
            new_value[_field.name] = _cast(value[key], _field.type)

    # noinspection PyArgumentList
    return to_type(**new_value)


def cast_to_type(value, to_type: type):
    if value is None:
        return None
    if issubclass(to_type, BaseModel):
        if not isinstance(value, dict):
            raise ValueError
        return cast_to_model(value, to_type)
    return to_type(value)


def _cast(value, to_type: type | list[type] | types.UnionType):
    to_type = unify_type(to_type)

    if to_type == dict:
        if isinstance(value, BaseModel):
            return asdict(value, dict_factory=lambda x: {k: v for (k, v) in x if v is not None})
        raise ValueError

    if isinstance(to_type, type):
        if is_generic_type(to_type):
            if get_type_origin(to_type) == list:
                subtypes = get_type_args(to_type)
                return [_cast(i, subtypes) for i in value]

        return cast_to_type(value, to_type)

    for subtype in to_type:
        try:
            return _cast(value, subtype)
        except (ValueError, TypeError):
            pass
    else:
        raise ValueError


def cast(value, to_type: type[T] | types.UnionType) -> T:
    try:
        return _cast(value, to_type)
    except ValueError:
        raise ValueError(f'Can\'t cast {repr(value)} to {to_type}')



from .answer_callback_query import answer_callback_query
from .copy_message import copy_message
from .delete_message import delete_message
from .delete_my_commands import delete_my_commands
from .edit_message_caption import edit_message_caption
from .edit_message_text import edit_message_text
from .get_chat_member import get_chat_member
from .get_file import get_file, get_file_url
from .get_me import get_me
from .send_chat_action import send_chat_action
from .send_media_group import send_media_group
from .send_message import send_message
from .send_photo import send_photo
from .send_video import send_video
from .set_my_commands import set_my_commands
from .get_chat import get_chat
from .edit_message_reply_markup import edit_message_reply_markup



from __future__ import annotations

from collections import defaultdict

from .context_dict import ContextDict
from .layer2 import ContextLayer2
from .storage import Storage
from ..objects import CallbackButton


def storage_property(name: str):
    def fget(self: ContextLayer3):
        memory_storage = self._memory_storage

        if name not in memory_storage:
            storage = self._storage
            value = getattr(storage, name)
            memory_storage[name] = value

        return memory_storage[name]

    def fset(self: ContextLayer3, value):
        storage = self._storage
        setattr(storage, name, value)
        storage.save()

        self._memory_storage[name] = value

    return property(fget, fset)


class ContextLayer3(ContextLayer2):
    def __init__(self):
        self.__memory_storage = defaultdict(dict)

    @property
    def _memory_storage(self):
        return self.__memory_storage[self.user_id]

    @property
    def _storage(self):
        return Storage.get(self.user_id)

    @property
    def data(self):
        return ContextDict(self._storage)

    lang: str | None = storage_property('lang')
    state: str | None = storage_property('state')

    @property
    def button(self) -> CallbackButton | None:
        """ CallbackQuery.button """
        try:
            doc_id = self.callback_query.data
            return CallbackButton.get(doc_id)
        except (KeyError, AttributeError):
            return None

    def delete_current_models(self):
        from ..base.user_proxy_model import CurrentObject

        for doc in CurrentObject.get_docs(user_id=self.user_id):
            doc.delete()



from dataclasses import dataclass

from .base import Filter
from ..context import ctx
from ..utils import listify


@dataclass
class State(Filter):
    value: str | list[str] = None

    def __call__(self):
        return self.value == '*' or ctx.state in listify(self.value)



from ..api import request
from ..context import ctx
from ..objects.tg_methods import DeleteMessage


def delete_message(
        message_id: int = None,
        chat_id: int | str = None,
) -> bool:
    return request(
        DeleteMessage,
        locals(),
        message_id=ctx.message_id,
        chat_id=ctx.chat_id,
    )



from ..api import request
from ..objects import Update
from ..objects.tg_methods import GetUpdates


def get_updates(
        offset: int = None,
        limit: int = None,
        timeout: int = None,
        allowed_updates: list[str] = None,
) -> list[Update]:
    return request(
        GetUpdates,
        locals(),
    )



from __future__ import annotations


class ConstantsMeta(type):

    def __getattribute__(cls: type[Constants], item):
        value = super().__getattribute__(item)

        if value is ...:
            if cls.__with_class_name__:
                return f'{cls.__name__}.{item}'
            return item

        return value


class Constants(metaclass=ConstantsMeta):
    __with_class_name__ = False



from ..api import request
from ..context import ctx
from ..objects.tg_methods import SendChatAction


def send_chat_action(
        action: str,
        chat_id: int | str = None,
) -> bool:
    return request(
        SendChatAction,
        locals(),
        chat_id=ctx.chat_id,
    )



from ..api import request
from ..context import ctx
from ..objects import Message, InputMediaDocument, InputMediaPhoto, InputMediaVideo, InputMediaAudio
from ..objects.tg_methods import SendMediaGroup


def send_media_group(
        media: list[InputMediaAudio | InputMediaDocument | InputMediaPhoto | InputMediaVideo],

        chat_id: int | str = None,
        disable_notification: bool = None,
        protect_content: bool = None,

        reply_to_message_id: int = None,
        allow_sending_without_reply: bool = None,
) -> list[Message]:
    _locals = locals()

    for m in media:
        if m.caption_entities:
            m.parse_mode = None
        else:
            m.parse_mode = m.parse_mode or ctx.parse_mode

    return request(
        SendMediaGroup,
        _locals,
        chat_id=ctx.chat_id,
        disable_notification=ctx.disable_notification,
        protect_content=ctx.protect_content,
    )



from ..api import request
from ..objects import BotCommand, BotCommandScope
from ..objects.tg_methods import AnswerCallbackQuery
from ..context import ctx


def answer_callback_query(
        text: str = None,
        show_alert: bool = None,

        callback_query_id: str = None,

        url: str = None,
        cache_time: int = None,
) -> bool:
    return request(
        AnswerCallbackQuery,
        locals(),
        callback_query_id=ctx.callback_query_id,
    )



from .new import (
    Translations,
    Keyboard,
    InlineKeyboard,
    CallbackButton,
    UrlButton,
)
from .tg_objects import (
    Update,
    WebhookInfo,
    User,
    Chat,
    ChatMemberT,
    Message,
    MessageId,
    MessageEntity,
    PhotoSize,
    Animation,
    Audio,
    Document,
    Video,
    VideoNote,
    Voice,
    Contact,
    Dice,
    PollOption,
    PollAnswer,
    Poll,
    Location,
    Venue,
    WebAppData,
    ProximityAlertTriggered,
    MessageAutoDeleteTimerChanged,
    VideoChatScheduled,
    VideoChatStarted,
    VideoChatEnded,
    VideoChatParticipantsInvited,
    UserProfilePhotos,
    File,
    WebAppInfo,
    ReplyKeyboardMarkup,
    KeyboardButton,
    KeyboardButtonPollType,
    ReplyKeyboardRemove,
    InlineKeyboardMarkup,
    InlineKeyboardButton,
    LoginUrl,
    CallbackQuery,
    ForceReply,
    ChatPhoto,
    ChatInviteLink,
    ChatAdministratorRights,
    ChatMember,
    ChatMemberOwner,
    ChatMemberAdministrator,
    ChatMemberMember,
    ChatMemberRestricted,
    ChatMemberLeft,
    ChatMemberBanned,
    ChatMemberUpdated,
    ChatJoinRequest,
    ChatPermissions,
    ChatLocation,
    BotCommand,
    BotCommandScope,
    BotCommandScopeDefault,
    BotCommandScopeAllPrivateChats,
    BotCommandScopeAllGroupChats,
    BotCommandScopeAllChatAdministrators,
    BotCommandScopeChat,
    BotCommandScopeChatAdministrators,
    BotCommandScopeChatMember,
    MenuButton,
    MenuButtonCommands,
    MenuButtonWebApp,
    MenuButtonDefault,
    ResponseParameters,
    InputMedia,
    InputMediaPhoto,
    InputMediaVideo,
    InputMediaAnimation,
    InputMediaAudio,
    InputMediaDocument,
    InputFile,
    Sticker,
    StickerSet,
    MaskPosition,
    InlineQuery,
    InlineQueryResult,
    InlineQueryResultArticle,
    InlineQueryResultPhoto,
    InlineQueryResultGif,
    InlineQueryResultMpeg4Gif,
    InlineQueryResultVideo,
    InlineQueryResultAudio,
    InlineQueryResultVoice,
    InlineQueryResultDocument,
    InlineQueryResultLocation,
    InlineQueryResultVenue,
    InlineQueryResultContact,
    InlineQueryResultGame,
    InlineQueryResultCachedPhoto,
    InlineQueryResultCachedGif,
    InlineQueryResultCachedMpeg4Gif,
    InlineQueryResultCachedSticker,
    InlineQueryResultCachedDocument,
    InlineQueryResultCachedVideo,
    InlineQueryResultCachedVoice,
    InlineQueryResultCachedAudio,
    InputMessageContent,
    InputTextMessageContent,
    InputLocationMessageContent,
    InputVenueMessageContent,
    InputContactMessageContent,
    InputInvoiceMessageContent,
    ChosenInlineResult,
    SentWebAppMessage,
    LabeledPrice,
    Invoice,
    ShippingAddress,
    OrderInfo,
    ShippingOption,
    SuccessfulPayment,
    ShippingQuery,
    PreCheckoutQuery,
    PassportData,
    PassportFile,
    EncryptedPassportElement,
    EncryptedCredentials,
    PassportElementError,
    PassportElementErrorDataField,
    PassportElementErrorFrontSide,
    PassportElementErrorReverseSide,
    PassportElementErrorSelfie,
    PassportElementErrorFile,
    PassportElementErrorFiles,
    PassportElementErrorTranslationFile,
    PassportElementErrorTranslationFiles,
    PassportElementErrorUnspecified,
    Game,
    CallbackGame,
    GameHighScore,
)



from .button import button
from .callback_query import callback_query
from .command import command
from .contact import contact
from .document import document
from .text import text
from .photo import photo
from .message import message
from .start import start
from .my_chat_member import my_chat_member



from ..objects.tg_objects import ReplyKeyboardMarkup, ReplyKeyboardRemove, InlineKeyboardMarkup, ForceReply
from ..objects.new import InlineKeyboard

ReplyMarkupT = ReplyKeyboardMarkup | InlineKeyboardMarkup | ReplyKeyboardRemove | ForceReply | InlineKeyboard



import typing

from requests import Response

from ..base import TgMethod, TgObject, BaseModel
from ..base.exceptions import RequestError
from ..loader import BOT_TOKEN, session
from ..objects import InlineKeyboard, Keyboard, Translations
from ..utils import cast, clean_dict

METHOD_URL = f'https://api.telegram.org/bot{BOT_TOKEN}/{{method}}'
FILE_URL = f'https://api.telegram.org/file/bot{BOT_TOKEN}/{{file_path}}'

OK = 'ok'
RESULT = 'result'
ERROR_CODE = 'error_code'
DESCRIPTION = 'description'


def get_method_url(method: str):
    return METHOD_URL.format(method=method)


def get_file_url(file_path: str):
    return FILE_URL.format(file_path=file_path)


def parse_response(resp: Response, method: type[TgMethod]):
    result: dict = resp.json()

    if result[OK]:
        result = result[RESULT]
    else:
        raise RequestError(result[ERROR_CODE], result[DESCRIPTION])

    return cast(result, method.__response_type__)


def request(method: type[TgMethod], params: dict, **alternatives) -> TgObject | typing.Any:
    endpoint = get_method_url(method.__name__)
    params = params.copy()

    for key in alternatives:
        if params[key] is None:
            params[key] = alternatives[key]

    for key, value in params.items():
        if isinstance(value, Keyboard):
            params[key] = value.to_tg_object()
        if isinstance(value, InlineKeyboard):
            params[key] = value.to_tg_object()
        if isinstance(value, Translations):
            params[key] = value.get()

    for key, value in params.items():
        if isinstance(value, BaseModel):
            params[key] = cast(value, dict)

    if params.get('entities') or params.get('caption_entities'):
        params['parse_mode'] = None

    params = clean_dict(params)
    resp = session.post(endpoint, json=params)
    return parse_response(resp, method)



import logging

import mongoengine
import requests
from envparse import env

from .handlers.base import Handler

BOT_TOKEN = env('BOT_TOKEN')
MONGO_DB = env('MONGO_DB')
MONGO_HOST = env('MONGO_HOST', 'localhost')

logger = logging.getLogger('bot')
mongoengine.connect(MONGO_DB, host=MONGO_HOST)
session = requests.Session()


HANDLERS: list[Handler] = []



from ..context import ctx
from ..api import request
from ..objects import ChatMemberT
from ..objects.tg_methods import GetChatMember


def get_chat_member(
        chat_id: int | str = None,
        user_id: int = None,
) -> ChatMemberT:
    return request(
        GetChatMember,
        locals(),
        chat_id=ctx.chat_id,
        user_id=ctx.user_id,
    )



from .message import Message, Text, Command, Contact, Document, Photo
from .state import State
from .user_id import UserId
from .chat_type import ChatType
from .callback_query import CallbackQuery
from .button import Button
from .my_chat_member import MyChatMember



from contextvars import ContextVar
from dataclasses import dataclass

from ..objects import Update

UPDATE = ContextVar('UPDATE')


@dataclass
class ContextLayer0:
    token: str = None
    parse_mode: str = None
    disable_web_page_preview: bool = None
    disable_notification: bool = None
    protect_content: bool = None

    @property
    def update(self) -> Update | None:
        """ Update """
        return UPDATE.get(None)

    @update.setter
    def update(self, value: Update):
        UPDATE.set(value)



import inspect
from threading import Thread
import yaml

from ..base.user_proxy_model import UserProxyModel
from ..base.exceptions import StopProcessing, ExitHandler
from ..context import ctx
from ..handlers.base import Handler
from ..loader import HANDLERS, logger
from ..objects.tg_objects import Update
from ..requests.get_updates import get_updates
from ..utils import cast


def execute_handler(handler: Handler):
    func = handler.func
    kwargs: dict[str, UserProxyModel] = {}

    for key, value in inspect.getfullargspec(func).annotations.items():
        value: type[UserProxyModel]
        kwargs[key] = value.get() or value()

    try:
        func(**kwargs)
    except ExitHandler:
        pass

    for model in kwargs.values():
        model.save()

    if handler.exclusive:
        raise StopProcessing()


def notify_first_handlers():
    for handler in HANDLERS:
        if not handler.check_first:
            continue

        for _filter in handler.filters:
            if not _filter():
                break
        else:
            execute_handler(handler)


def notify_default_handlers():
    for handler in HANDLERS:
        if handler.check_first or handler.check_last or handler.check_after_any:
            continue

        for _filter in handler.filters:
            if not _filter():
                break
        else:
            execute_handler(handler)


def notify_last_handlers():
    for handler in HANDLERS:
        if not handler.check_last:
            continue

        for _filter in handler.filters:
            if not _filter():
                break
        else:
            execute_handler(handler)


def notify_after_any_handlers():
    for handler in HANDLERS:
        if not handler.check_after_any:
            continue

        for _filter in handler.filters:
            if not _filter():
                break
        else:
            execute_handler(handler)


def notify_handlers():
    try:
        notify_first_handlers()
        notify_default_handlers()
        notify_last_handlers()
    except StopProcessing:
        pass

    try:
        notify_after_any_handlers()
    except StopProcessing:
        pass


def process_update(update: Update) -> None:
    ctx.update = update
    notify_handlers()
    ctx.update = None


def process_updates(updates: list[Update]):
    for update in updates:
        Thread(target=lambda: process_update(update)).start()

        as_dict = cast(update, dict)
        as_string = yaml.dump(as_dict, allow_unicode=True, sort_keys=False)
        as_string = f'Update ->\n{as_string}'
        logger.info(as_string)


def start_polling(skip_updates: bool):
    offset = None

    if skip_updates:
        updates = get_updates(offset=offset)
        if updates:
            offset = updates[-1].update_id + 1

    while True:
        try:
            updates = get_updates(offset=offset)

            if updates:
                offset = updates[-1].update_id + 1
                process_updates(updates)
        except Exception as exc:
            logger.exception(exc)


def run(
        skip_updates: bool = False,
        parse_mode: str = None,
        disable_web_page_preview: bool = None,
        disable_notification: bool = None,
        protect_content: bool = None,
):
    logger.info('Starting up...')

    ctx.parse_mode = parse_mode
    ctx.disable_web_page_preview = disable_web_page_preview
    ctx.disable_notification = disable_notification
    ctx.protect_content = protect_content

    try:
        start_polling(skip_updates)
    except KeyboardInterrupt:
        logger.info('Shutting down...')



from . import handlers as on
from . import objects as objects
from . import requests as bot
from .base import UserProxyModel, Constants, exceptions as exc
from .context import ctx
from .loader import logger
from .run import run
from .utils import html

__all__ = ['on', 'objects', 'bot', 'UserProxyModel', 'Constants', 'ctx', 'run', 'exc', 'logger', 'html']



from dataclasses import dataclass

from .base import Filter
from ..context import ctx
from ..utils import listify


@dataclass
class CallbackQuery(Filter):
    data: str | list[str] = None

    def __call__(self):
        try:
            if not (data := ctx.update.callback_query.data):
                return False
        except AttributeError:
            return False

        if self.data is None:
            return True

        return data in listify(self.data)



from .callback_button import CallbackButton
from .url_button import UrlButton
from ...base import NewObject
from ..tg_objects import InlineKeyboardMarkup, InlineKeyboardButton
from ..new.translations import Translations

BUTTONS = '__buttons'


class InlineKeyboard(NewObject):

    def __new__(cls, *args, **kwargs):
        kb = super().__new__(cls)
        kb._buttons = []
        return kb

    def add_row(self, *buttons: CallbackButton | UrlButton):
        self._buttons.append(list(buttons))

    def add_rows(self, *buttons: CallbackButton | UrlButton):
        self._buttons.extend([btn] for btn in buttons)

    def __repr__(self):
        return f'{InlineKeyboard.__name__}({self._buttons})'

    def to_tg_object(self) -> InlineKeyboardMarkup:
        new_buttons: list[list[InlineKeyboardButton]] = []

        for row in self._buttons:
            new_row = []
            for btn in row:

                text = btn.text
                if isinstance(text, Translations):
                    text = text.get()

                if isinstance(btn, CallbackButton):
                    new_btn = InlineKeyboardButton(text, callback_data=btn.doc_id)
                elif isinstance(btn, UrlButton):
                    new_btn = InlineKeyboardButton(text, url=btn.url)
                else:
                    raise ValueError('Invalid button')

                new_row.append(new_btn)

            new_buttons.append(new_row)

        return InlineKeyboardMarkup(new_buttons)



""" All telegram requests from https://core.telegram.org/bots/api (v 6.0) """

from __future__ import annotations

from .tg_objects import *
from ..base import TgMethod


# ==> Section: https://core.telegram.org/bots/api#getting-updates

@dataclass
class GetUpdates(TgMethod):
    """ https://core.telegram.org/bots/api#getupdates """

    offset: int = None
    limit: int = None
    timeout: int = None
    allowed_updates: list[str] = None

    __response_type__ = list[Update]


@dataclass
class SetWebhook(TgMethod):
    """ https://core.telegram.org/bots/api#setwebhook """

    url: str
    certificate: InputFile = None
    ip_address: str = None
    max_connections: int = None
    allowed_updates: list[str] = None
    drop_pending_updates: bool = None

    __response_type__ = bool


@dataclass
class DeleteWebhook(TgMethod):
    """ https://core.telegram.org/bots/api#deletewebhook """

    drop_pending_updates: bool = None

    __response_type__ = bool


@dataclass
class GetWebhookInfo(TgMethod):
    """ https://core.telegram.org/bots/api#getwebhookinfo """

    __response_type__ = WebhookInfo


# ==> Section: https://core.telegram.org/bots/api#available-methods

@dataclass
class GetMe(TgMethod):
    """ https://core.telegram.org/bots/api#getme """

    __response_type__ = User


@dataclass
class LogOut(TgMethod):
    """ https://core.telegram.org/bots/api#logout """

    __response_type__ = bool


@dataclass
class Close(TgMethod):
    """ https://core.telegram.org/bots/api#close """

    __response_type__ = bool


@dataclass
class SendMessage(TgMethod):
    """ https://core.telegram.org/bots/api#sendmessage """

    chat_id: int | str
    text: str
    parse_mode: str = None
    entities: list[MessageEntity] = None
    disable_web_page_preview: bool = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class ForwardMessage(TgMethod):
    """ https://core.telegram.org/bots/api#forwardmessage """

    chat_id: int | str
    from_chat_id: int | str
    message_id: int
    disable_notification: bool = None
    protect_content: bool = None

    __response_type__ = Message


@dataclass
class CopyMessage(TgMethod):
    """ https://core.telegram.org/bots/api#copymessage """

    chat_id: int | str
    from_chat_id: int | str
    message_id: int
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = MessageId


@dataclass
class SendPhoto(TgMethod):
    """ https://core.telegram.org/bots/api#sendphoto """

    chat_id: int | str
    photo: InputFile | str
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class SendAudio(TgMethod):
    """ https://core.telegram.org/bots/api#sendaudio """

    chat_id: int | str
    audio: InputFile | str
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    duration: int = None
    performer: str = None
    title: str = None
    thumb: InputFile | str = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class SendDocument(TgMethod):
    """ https://core.telegram.org/bots/api#senddocument """

    chat_id: int | str
    document: InputFile | str
    thumb: InputFile | str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    disable_content_type_detection: bool = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class SendVideo(TgMethod):
    """ https://core.telegram.org/bots/api#sendvideo """

    chat_id: int | str
    video: InputFile | str
    duration: int = None
    width: int = None
    height: int = None
    thumb: InputFile | str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    supports_streaming: bool = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class SendAnimation(TgMethod):
    """ https://core.telegram.org/bots/api#sendanimation """

    chat_id: int | str
    animation: InputFile | str
    duration: int = None
    width: int = None
    height: int = None
    thumb: InputFile | str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class SendVoice(TgMethod):
    """ https://core.telegram.org/bots/api#sendvoice """

    chat_id: int | str
    voice: InputFile | str
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    duration: int = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class SendVideoNote(TgMethod):
    """ https://core.telegram.org/bots/api#sendvideonote """

    chat_id: int | str
    video_note: InputFile | str
    duration: int = None
    length: int = None
    thumb: InputFile | str = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class SendMediaGroup(TgMethod):
    """ https://core.telegram.org/bots/api#sendmediagroup """

    chat_id: int | str
    media: list[InputMediaAudio | InputMediaDocument | InputMediaPhoto | InputMediaVideo]
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None

    __response_type__ = list[Message]


@dataclass
class SendLocation(TgMethod):
    """ https://core.telegram.org/bots/api#sendlocation """

    chat_id: int | str
    latitude: float
    longitude: float
    horizontal_accuracy: float = None
    live_period: int = None
    heading: int = None
    proximity_alert_radius: int = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class EditMessageLiveLocation(TgMethod):
    """ https://core.telegram.org/bots/api#editmessagelivelocation """

    latitude: float
    longitude: float
    chat_id: int | str = None
    message_id: int = None
    inline_message_id: str = None
    horizontal_accuracy: float = None
    heading: int = None
    proximity_alert_radius: int = None
    reply_markup: InlineKeyboardMarkup = None

    __response_type__ = Message | bool


@dataclass
class StopMessageLiveLocation(TgMethod):
    """ https://core.telegram.org/bots/api#stopmessagelivelocation """

    chat_id: int | str = None
    message_id: int = None
    inline_message_id: str = None
    reply_markup: InlineKeyboardMarkup = None

    __response_type__ = Message | bool


@dataclass
class SendVenue(TgMethod):
    """ https://core.telegram.org/bots/api#sendvenue """

    chat_id: int | str
    latitude: float
    longitude: float
    title: str
    address: str
    foursquare_id: str = None
    foursquare_type: str = None
    google_place_id: str = None
    google_place_type: str = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class SendContact(TgMethod):
    """ https://core.telegram.org/bots/api#sendcontact """

    chat_id: int | str
    phone_number: str
    first_name: str
    last_name: str = None
    vcard: str = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class SendPoll(TgMethod):
    """ https://core.telegram.org/bots/api#sendpoll """

    chat_id: int | str
    question: str
    options: list[str]
    is_anonymous: bool = None
    type: str = None
    allows_multiple_answers: bool = None
    correct_option_id: int = None
    explanation: str = None
    explanation_parse_mode: str = None
    explanation_entities: list[MessageEntity] = None
    open_period: int = None
    close_date: int = None
    is_closed: bool = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class SendDice(TgMethod):
    """ https://core.telegram.org/bots/api#senddice """

    chat_id: int | str
    emoji: str = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class SendChatAction(TgMethod):
    """ https://core.telegram.org/bots/api#sendchataction """

    chat_id: int | str
    action: str

    __response_type__ = bool


@dataclass
class GetUserProfilePhotos(TgMethod):
    """ https://core.telegram.org/bots/api#getuserprofilephotos """

    user_id: int
    offset: int = None
    limit: int = None

    __response_type__ = UserProfilePhotos


@dataclass
class GetFile(TgMethod):
    """ https://core.telegram.org/bots/api#getfile """

    file_id: str

    __response_type__ = File


@dataclass
class BanChatMember(TgMethod):
    """ https://core.telegram.org/bots/api#banchatmember """

    chat_id: int | str
    user_id: int
    until_date: int = None
    revoke_messages: bool = None

    __response_type__ = bool


@dataclass
class UnbanChatMember(TgMethod):
    """ https://core.telegram.org/bots/api#unbanchatmember """

    chat_id: int | str
    user_id: int
    only_if_banned: bool = None

    __response_type__ = bool


@dataclass
class RestrictChatMember(TgMethod):
    """ https://core.telegram.org/bots/api#restrictchatmember """

    chat_id: int | str
    user_id: int
    permissions: ChatPermissions
    until_date: int = None

    __response_type__ = bool


@dataclass
class PromoteChatMember(TgMethod):
    """ https://core.telegram.org/bots/api#promotechatmember """

    chat_id: int | str
    user_id: int
    is_anonymous: bool = None
    can_manage_chat: bool = None
    can_post_messages: bool = None
    can_edit_messages: bool = None
    can_delete_messages: bool = None
    can_manage_video_chats: bool = None
    can_restrict_members: bool = None
    can_promote_members: bool = None
    can_change_info: bool = None
    can_invite_users: bool = None
    can_pin_messages: bool = None

    __response_type__ = bool


@dataclass
class SetChatAdministratorCustomTitle(TgMethod):
    """ https://core.telegram.org/bots/api#setchatadministratorcustomtitle """

    chat_id: int | str
    user_id: int
    custom_title: str

    __response_type__ = bool


@dataclass
class BanChatSenderChat(TgMethod):
    """ https://core.telegram.org/bots/api#banchatsenderchat """

    chat_id: int | str
    sender_chat_id: int

    __response_type__ = bool


@dataclass
class UnbanChatSenderChat(TgMethod):
    """ https://core.telegram.org/bots/api#unbanchatsenderchat """

    chat_id: int | str
    sender_chat_id: int

    __response_type__ = bool


@dataclass
class SetChatPermissions(TgMethod):
    """ https://core.telegram.org/bots/api#setchatpermissions """

    chat_id: int | str
    permissions: ChatPermissions

    __response_type__ = bool


@dataclass
class ExportChatInviteLink(TgMethod):
    """ https://core.telegram.org/bots/api#exportchatinvitelink """

    chat_id: int | str

    __response_type__ = str


@dataclass
class CreateChatInviteLink(TgMethod):
    """ https://core.telegram.org/bots/api#createchatinvitelink """

    chat_id: int | str
    name: str = None
    expire_date: int = None
    member_limit: int = None
    creates_join_request: bool = None

    __response_type__ = ChatInviteLink


@dataclass
class EditChatInviteLink(TgMethod):
    """ https://core.telegram.org/bots/api#editchatinvitelink """

    chat_id: int | str
    invite_link: str
    name: str = None
    expire_date: int = None
    member_limit: int = None
    creates_join_request: bool = None

    __response_type__ = ChatInviteLink


@dataclass
class RevokeChatInviteLink(TgMethod):
    """ https://core.telegram.org/bots/api#revokechatinvitelink """

    chat_id: int | str
    invite_link: str

    __response_type__ = ChatInviteLink


@dataclass
class ApproveChatJoinRequest(TgMethod):
    """ https://core.telegram.org/bots/api#approvechatjoinrequest """

    chat_id: int | str
    user_id: int

    __response_type__ = bool


@dataclass
class DeclineChatJoinRequest(TgMethod):
    """ https://core.telegram.org/bots/api#declinechatjoinrequest """

    chat_id: int | str
    user_id: int

    __response_type__ = bool


@dataclass
class SetChatPhoto(TgMethod):
    """ https://core.telegram.org/bots/api#setchatphoto """

    chat_id: int | str
    photo: InputFile

    __response_type__ = bool


@dataclass
class DeleteChatPhoto(TgMethod):
    """ https://core.telegram.org/bots/api#deletechatphoto """

    chat_id: int | str

    __response_type__ = bool


@dataclass
class SetChatTitle(TgMethod):
    """ https://core.telegram.org/bots/api#setchattitle """

    chat_id: int | str
    title: str

    __response_type__ = bool


@dataclass
class SetChatDescription(TgMethod):
    """ https://core.telegram.org/bots/api#setchatdescription """

    chat_id: int | str
    description: str = None

    __response_type__ = bool


@dataclass
class PinChatMessage(TgMethod):
    """ https://core.telegram.org/bots/api#pinchatmessage """

    chat_id: int | str
    message_id: int
    disable_notification: bool = None

    __response_type__ = bool


@dataclass
class UnpinChatMessage(TgMethod):
    """ https://core.telegram.org/bots/api#unpinchatmessage """

    chat_id: int | str
    message_id: int = None

    __response_type__ = bool


@dataclass
class UnpinAllChatMessages(TgMethod):
    """ https://core.telegram.org/bots/api#unpinallchatmessages """

    chat_id: int | str

    __response_type__ = bool


@dataclass
class LeaveChat(TgMethod):
    """ https://core.telegram.org/bots/api#leavechat """

    chat_id: int | str

    __response_type__ = bool


@dataclass
class GetChat(TgMethod):
    """ https://core.telegram.org/bots/api#getchat """

    chat_id: int | str

    __response_type__ = Chat


@dataclass
class GetChatAdministrators(TgMethod):
    """ https://core.telegram.org/bots/api#getchatadministrators """

    chat_id: int | str

    __response_type__ = list[ChatMember]


@dataclass
class GetChatMemberCount(TgMethod):
    """ https://core.telegram.org/bots/api#getchatmembercount """

    chat_id: int | str

    __response_type__ = int


@dataclass
class GetChatMember(TgMethod):
    """ https://core.telegram.org/bots/api#getchatmember """

    chat_id: int | str
    user_id: int

    __response_type__ = ChatMemberT


@dataclass
class SetChatStickerSet(TgMethod):
    """ https://core.telegram.org/bots/api#setchatstickerset """

    chat_id: int | str
    sticker_set_name: str

    __response_type__ = bool


@dataclass
class DeleteChatStickerSet(TgMethod):
    """ https://core.telegram.org/bots/api#deletechatstickerset """

    chat_id: int | str

    __response_type__ = bool


@dataclass
class AnswerCallbackQuery(TgMethod):
    """ https://core.telegram.org/bots/api#answercallbackquery """

    callback_query_id: str
    text: str = None
    show_alert: bool = None
    url: str = None
    cache_time: int = None

    __response_type__ = bool


@dataclass
class SetMyCommands(TgMethod):
    """ https://core.telegram.org/bots/api#setmycommands """

    commands: list[BotCommand]
    scope: BotCommandScope = None
    language_code: str = None

    __response_type__ = bool


@dataclass
class DeleteMyCommands(TgMethod):
    """ https://core.telegram.org/bots/api#deletemycommands """

    scope: BotCommandScope = None
    language_code: str = None

    __response_type__ = bool


@dataclass
class GetMyCommands(TgMethod):
    """ https://core.telegram.org/bots/api#getmycommands """

    scope: BotCommandScope = None
    language_code: str = None

    __response_type__ = list[BotCommand]


@dataclass
class SetChatMenuButton(TgMethod):
    """ https://core.telegram.org/bots/api#setchatmenubutton """

    chat_id: int = None
    menu_button: MenuButton = None

    __response_type__ = bool


@dataclass
class GetChatMenuButton(TgMethod):
    """ https://core.telegram.org/bots/api#getchatmenubutton """

    chat_id: int = None

    __response_type__ = MenuButton


@dataclass
class SetMyDefaultAdministratorRights(TgMethod):
    """ https://core.telegram.org/bots/api#setmydefaultadministratorrights """

    rights: ChatAdministratorRights = None
    for_channels: bool = None

    __response_type__ = bool


@dataclass
class GetMyDefaultAdministratorRights(TgMethod):
    """ https://core.telegram.org/bots/api#getmydefaultadministratorrights """

    for_channels: bool = None

    __response_type__ = ChatAdministratorRights


# ==> Section: https://core.telegram.org/bots/api#updating-messages

@dataclass
class EditMessageText(TgMethod):
    """ https://core.telegram.org/bots/api#editmessagetext """

    text: str
    chat_id: int | str = None
    message_id: int = None
    inline_message_id: str = None
    parse_mode: str = None
    entities: list[MessageEntity] = None
    disable_web_page_preview: bool = None
    reply_markup: InlineKeyboardMarkup = None

    __response_type__ = Message | bool


@dataclass
class EditMessageCaption(TgMethod):
    """ https://core.telegram.org/bots/api#editmessagecaption """

    chat_id: int | str = None
    message_id: int = None
    inline_message_id: str = None
    caption: str = None
    parse_mode: str = None
    caption_entities: list[MessageEntity] = None
    reply_markup: InlineKeyboardMarkup = None

    __response_type__ = Message | bool


@dataclass
class EditMessageMedia(TgMethod):
    """ https://core.telegram.org/bots/api#editmessagemedia """

    media: InputMedia
    chat_id: int | str = None
    message_id: int = None
    inline_message_id: str = None
    reply_markup: InlineKeyboardMarkup = None

    __response_type__ = Message | bool


@dataclass
class EditMessageReplyMarkup(TgMethod):
    """ https://core.telegram.org/bots/api#editmessagereplymarkup """

    chat_id: int | str = None
    message_id: int = None
    inline_message_id: str = None
    reply_markup: InlineKeyboardMarkup = None

    __response_type__ = Message | bool


@dataclass
class StopPoll(TgMethod):
    """ https://core.telegram.org/bots/api#stoppoll """

    chat_id: int | str
    message_id: int
    reply_markup: InlineKeyboardMarkup = None

    __response_type__ = Poll


@dataclass
class DeleteMessage(TgMethod):
    """ https://core.telegram.org/bots/api#deletemessage """

    chat_id: int | str
    message_id: int

    __response_type__ = bool


# ==> Section: https://core.telegram.org/bots/api#stickers

@dataclass
class SendSticker(TgMethod):
    """ https://core.telegram.org/bots/api#sendsticker """

    chat_id: int | str
    sticker: InputFile | str
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup | ReplyKeyboardMarkup | ReplyKeyboardRemove | ForceReply = None

    __response_type__ = Message


@dataclass
class GetStickerSet(TgMethod):
    """ https://core.telegram.org/bots/api#getstickerset """

    name: str

    __response_type__ = StickerSet


@dataclass
class UploadStickerFile(TgMethod):
    """ https://core.telegram.org/bots/api#uploadstickerfile """

    user_id: int
    png_sticker: InputFile

    __response_type__ = File


@dataclass
class CreateNewStickerSet(TgMethod):
    """ https://core.telegram.org/bots/api#createnewstickerset """

    user_id: int
    name: str
    title: str
    emojis: str
    png_sticker: InputFile | str = None
    tgs_sticker: InputFile = None
    webm_sticker: InputFile = None
    contains_masks: bool = None
    mask_position: MaskPosition = None

    __response_type__ = bool


@dataclass
class AddStickerToSet(TgMethod):
    """ https://core.telegram.org/bots/api#addstickertoset """

    user_id: int
    name: str
    emojis: str
    png_sticker: InputFile | str = None
    tgs_sticker: InputFile = None
    webm_sticker: InputFile = None
    mask_position: MaskPosition = None

    __response_type__ = bool


@dataclass
class SetStickerPositionInSet(TgMethod):
    """ https://core.telegram.org/bots/api#setstickerpositioninset """

    sticker: str
    position: int

    __response_type__ = bool


@dataclass
class DeleteStickerFromSet(TgMethod):
    """ https://core.telegram.org/bots/api#deletestickerfromset """

    sticker: str

    __response_type__ = bool


@dataclass
class SetStickerSetThumb(TgMethod):
    """ https://core.telegram.org/bots/api#setstickersetthumb """

    name: str
    user_id: int
    thumb: InputFile | str = None

    __response_type__ = bool


# ==> Section: https://core.telegram.org/bots/api#inline-mode

@dataclass
class AnswerInlineQuery(TgMethod):
    """ https://core.telegram.org/bots/api#answerinlinequery """

    inline_query_id: str
    results: list[InlineQueryResult]
    cache_time: int = None
    is_personal: bool = None
    next_offset: str = None
    switch_pm_text: str = None
    switch_pm_parameter: str = None

    __response_type__ = bool


@dataclass
class AnswerWebAppQuery(TgMethod):
    """ https://core.telegram.org/bots/api#answerwebappquery """

    web_app_query_id: str
    result: InlineQueryResult

    __response_type__ = SentWebAppMessage


# ==> Section: https://core.telegram.org/bots/api#payments

@dataclass
class SendInvoice(TgMethod):
    """ https://core.telegram.org/bots/api#sendinvoice """

    chat_id: int | str
    title: str
    description: str
    payload: str
    provider_token: str
    currency: str
    prices: list[LabeledPrice]
    max_tip_amount: int = None
    suggested_tip_amounts: list[int] = None
    start_parameter: str = None
    provider_data: str = None
    photo_url: str = None
    photo_size: int = None
    photo_width: int = None
    photo_height: int = None
    need_name: bool = None
    need_phone_number: bool = None
    need_email: bool = None
    need_shipping_address: bool = None
    send_phone_number_to_provider: bool = None
    send_email_to_provider: bool = None
    is_flexible: bool = None
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup = None

    __response_type__ = Message


@dataclass
class AnswerShippingQuery(TgMethod):
    """ https://core.telegram.org/bots/api#answershippingquery """

    shipping_query_id: str
    ok: bool
    shipping_options: list[ShippingOption] = None
    error_message: str = None

    __response_type__ = bool


@dataclass
class AnswerPreCheckoutQuery(TgMethod):
    """ https://core.telegram.org/bots/api#answerprecheckoutquery """

    pre_checkout_query_id: str
    ok: bool
    error_message: str = None

    __response_type__ = bool


# ==> Section: https://core.telegram.org/bots/api#telegram-passport

@dataclass
class SetPassportDataErrors(TgMethod):
    """ https://core.telegram.org/bots/api#setpassportdataerrors """

    user_id: int
    errors: list[PassportElementError]

    __response_type__ = bool


# ==> Section: https://core.telegram.org/bots/api#games

@dataclass
class SendGame(TgMethod):
    """ https://core.telegram.org/bots/api#sendgame """

    chat_id: int
    game_short_name: str
    disable_notification: bool = None
    protect_content: bool = None
    reply_to_message_id: int = None
    allow_sending_without_reply: bool = None
    reply_markup: InlineKeyboardMarkup = None

    __response_type__ = Message


@dataclass
class SetGameScore(TgMethod):
    """ https://core.telegram.org/bots/api#setgamescore """

    user_id: int
    score: int
    force: bool = None
    disable_edit_message: bool = None
    chat_id: int = None
    message_id: int = None
    inline_message_id: str = None

    __response_type__ = Message | bool


@dataclass
class GetGameHighScores(TgMethod):
    """ https://core.telegram.org/bots/api#getgamehighscores """

    user_id: int
    chat_id: int = None
    message_id: int = None
    inline_message_id: str = None

    __response_type__ = list[GameHighScore]



from ..api import request
from ..context import ctx
from ..objects import InlineKeyboard, InlineKeyboardMarkup, MessageEntity, Message
from ..objects.tg_methods import EditMessageCaption


def edit_message_caption(
        caption: str = None,
        reply_markup: InlineKeyboard | InlineKeyboardMarkup = None,

        chat_id: int | str = None,
        message_id: int = None,
        inline_message_id: str = None,
        parse_mode: str = None,

        caption_entities: list[MessageEntity] = None,
) -> Message | bool:
    return request(
        EditMessageCaption,
        locals(),
        chat_id=ctx.chat_id,
        message_id=ctx.message_id,
        parse_mode=ctx.parse_mode,
    )



from dataclasses import dataclass

from .storage import Storage


@dataclass
class ContextDict:
    storage: Storage

    def __getitem__(self, item: str):
        return self.storage.data[item]

    def __setitem__(self, key: str, value):
        self.storage.data[key] = value
        self.storage.save()

    def __delitem__(self, key: str):
        del self.storage.data[key]
        self.storage.save()

    def __contains__(self, key: str):
        return key in self.storage.data

    def get(self, item: str, default=None):
        try:
            return self[item]
        except KeyError:
            return default

    def setdefault(self, key: str, default):
        self.storage.data.setdefault(key, default)
        self.storage.save()

    def clear(self):
        self.storage.data.clear()
        self.storage.save()



from ..api import request
from ..context import ctx
from ..objects import Translations, MessageEntity, Message, InlineKeyboardMarkup, InlineKeyboard
from ..objects.tg_methods import EditMessageText


def edit_message_text(
        text: str | Translations,
        reply_markup: InlineKeyboardMarkup | InlineKeyboard = None,

        chat_id: int | str = None,
        message_id: int = None,
        inline_message_id: str = None,
        parse_mode: str = None,
        disable_web_page_preview: bool = None,

        entities: list[MessageEntity] = None,
) -> Message | bool:
    return request(
        EditMessageText,
        locals(),
        chat_id=ctx.chat_id,
        message_id=ctx.message_id,
        parse_mode=ctx.parse_mode,
        disable_web_page_preview=ctx.disable_web_page_preview,
    )



from .command import command


def start(
        user_id: int | list[int] = None,
        chat_type: str | list[str] = None,
        state: str = None,
):
    return command('start', user_id, chat_type, state)



from dataclasses import dataclass

from .base import Filter
from ..context import ctx


@dataclass
class MyChatMember(Filter):
    def __call__(self):
        try:
            return bool(ctx.update.my_chat_member)
        except AttributeError:
            return False



from dataclasses import dataclass


@dataclass
class Error(Exception):
    def __str__(self):
        return repr(self)


@dataclass
class RequestError(Error):
    error_code: int
    description: str


class ExitHandler(Exception):
    pass


class StopProcessing(Exception):
    pass



from __future__ import annotations

from dataclasses import dataclass

from ...base import NewObject


@dataclass
class UrlButton(NewObject):
    """ UrlButton """
    text: str
    url: str = None

    def __call__(self, **_vars) -> UrlButton:
        """ Return button copy [text=text.format(**_vars), url=url.format(**_vars)] """
        return UrlButton(self.text.format(**_vars), self.url.format(**_vars))



from .base import Handler
from .. import filters
from ..loader import HANDLERS


def message(
        user_id: int | list[int] = None,
        chat_type: str | list[str] = None,
        state: str = None,
):
    _filters = [
        filters.Message(),
        filters.State(state),
    ]

    if user_id:
        _filters.append(filters.UserId(user_id))

    if chat_type:
        _filters.append(filters.ChatType(chat_type))

    def _(func):
        if isinstance(func, Handler):
            func = func.func

        handler = Handler(func, _filters)
        HANDLERS.append(handler)
        return handler

    return _



from dataclasses import dataclass

from .base import Filter
from ..context import ctx
from ..utils import listify


@dataclass
class ChatType(Filter):
    value: str | list[str]

    def __call__(self):
        return ctx.chat_type in listify(self.value)



import typing

T = typing.TypeVar('T')


def listify(obj: T, cast_tuple=True) -> T | list[T]:
    if cast_tuple:
        if isinstance(obj, tuple):
            return list(obj)

    if isinstance(obj, list):
        return obj

    return [obj]



from __future__ import annotations

import mongoengine as me

from ..base import BaseDocument


class Storage(BaseDocument):
    key: int = me.IntField()
    state: str = me.StringField()
    lang: str = me.StringField()
    data: dict = me.DictField()

    @classmethod
    def get(cls, user_id: int) -> Storage:
        return cls.get_doc(key=user_id) or Storage(key=user_id).save()

    meta = {
        'collection': 'Storage'
    }



import json
import typing

import mongoengine as me

DocT = typing.TypeVar('DocT', bound='BaseDocument')


class BaseDocument(me.Document):
    meta = {
        'abstract': True,
    }

    @classmethod
    def get_doc(cls: type[DocT], *args, **kwargs) -> DocT | None:
        return cls.objects(*args, **kwargs).first()

    @classmethod
    def get_docs(cls: type[DocT], *args, **kwargs) -> list[DocT]:
        return list(cls.objects(*args, **kwargs))

    def to_dict(self) -> dict:
        return json.loads(self.to_json())



from copy import deepcopy

from ...base import NewObject
from ..tg_objects import KeyboardButton, ReplyKeyboardMarkup
from .translations import Translations


class Keyboard(NewObject):

    def __new__(cls, *args, **kwargs):
        kb = super().__new__(cls)
        kb._buttons = []
        return kb

    def add_row(self, *buttons: KeyboardButton | str):
        self._buttons.append(list(buttons))

    def add_rows(self, *buttons: KeyboardButton | str, row_width: int = 1):
        index = 0

        while index < len(buttons):
            row = buttons[index:index + row_width]
            self._buttons.append(list(row))
            index += row_width

    def __repr__(self):
        return f'{Keyboard.__name__}({self._buttons})'

    def to_tg_object(self) -> ReplyKeyboardMarkup:
        new_buttons: list[list[KeyboardButton]] = []

        for row in self._buttons:
            new_row = []
            for btn in row:
                if isinstance(btn, (str, Translations)):
                    btn = KeyboardButton(btn)
                elif isinstance(btn, KeyboardButton):
                    btn = deepcopy(btn)
                else:
                    raise ValueError('Invalid button')

                if isinstance(btn.text, Translations):
                    btn.text = btn.text.get()

                new_row.append(btn)

            new_buttons.append(new_row)

        return ReplyKeyboardMarkup(new_buttons, resize_keyboard=True)



from ...base import NewObject
from dataclasses import dataclass, astuple


@dataclass
class Translations(NewObject):
    """ Translations """

    def __hash__(self):
        return hash(astuple(self))

    def get(self, lang: str = None) -> str | None:
        from ...context import ctx

        return getattr(self, lang or ctx.lang, None)

    def format(self, *args, **kwargs):
        # noinspection PyArgumentList
        return self.__class__(*[i.format(*args, **kwargs) for i in astuple(self)])

    def __contains__(self, item):
        return item in astuple(self)



from ..api import request
from ..context import ctx
from ..objects import Message, InlineKeyboardMarkup, InlineKeyboard
from ..objects.tg_methods import EditMessageReplyMarkup


def edit_message_reply_markup(
        reply_markup: InlineKeyboardMarkup | InlineKeyboard = None,
        chat_id: int | str = None,
        message_id: int = None,
        inline_message_id: str = None,
) -> Message | bool:
    return request(
        EditMessageReplyMarkup,
        locals(),
        chat_id=ctx.chat_id,
        message_id=ctx.message_id,
    )



from dataclasses import dataclass

from .base import Filter
from ..context import ctx
from ..utils import listify


@dataclass
class Message(Filter):
    def __call__(self):
        try:
            return bool(ctx.update.message)
        except AttributeError:
            return False


@dataclass
class Text(Filter):
    value: str | list[str] = None

    def __call__(self):
        try:
            if not (text := ctx.update.message.text):
                return False
        except AttributeError:
            return False

        if self.value is None:
            return bool(text)
        return text in listify(self.value)


@dataclass
class Command(Filter):
    value: str | list[str] = None

    def __call__(self):
        try:
            if not (text := ctx.update.message.text):
                return False
        except AttributeError:
            return False

        if self.value is None:
            return text.startswith('/')

        return text.lstrip('/') in listify(self.value)


@dataclass
class Contact(Filter):
    def __call__(self):
        try:
            return bool(ctx.update.message.contact)
        except AttributeError:
            return False


@dataclass
class Document(Filter):
    def __call__(self):
        try:
            return bool(ctx.update.message.document)
        except AttributeError:
            return False


@dataclass
class Photo(Filter):
    def __call__(self):
        try:
            return bool(ctx.update.message.photo)
        except AttributeError:
            return False
